function [ results ] = run( k, lambda, nTrain, nTest, nChars, kernels, varargin )
%RUN Train and test SVMs using the given string kernels
%   k - number of chars in substring
%   lambda - gap penalty
%   nTrain - vector of number of examples of each class to learn
%   nTest - vector of number of samples of each class to classify
%   nChars - maximum width of input strings
%   kernels - cell array of kernel names
%   (theta) - lamda pruning approximation factor
%
%   Example:
%      RUN(4, .5, [15 15], [10 10], 400, {'SSK','SSK-LP','SSK-LPC'},
%      3)
%      will report results for SSK, SSK-LP (lambda pruning), and SSK-LPC
%      (cached lambda pruning)
%
%   Returns a structure giving the results of the tests


p = inputParser;

checkLambda = @(x) isnumeric(x) && 0 <= x && x <= 1;
checkK = @(k) isnumeric(k) && 1 <= k;
addRequired(p,'k',checkK)
addRequired(p,'lambda',checkLambda)
addRequired(p,'nTrain',@isnumeric);
addRequired(p,'nTest',@isnumeric);
addRequired(p,'nChars',@isnumeric);
addRequired(p,'kernel',@(x)iscell(x) || ischar(x));
addOptional(p,'theta',@isnumeric);
parse(p,k, lambda, nTrain, nTest, nChars, kernels, varargin{:});

% Get data
[trainData classes] = getdata('train.txt', nTrain, true);
testData = getdata('test.txt', nTest, true);

% Convert to uint8 matrix
[training group testing answer] = formatinputs(trainData, testData, classes, nChars);


if ~iscell(kernels)
    kernels = {kernels};
end

results = cell(1, size(kernels, 2));

if length(varargin) < 1
    theta = 3;
else
    theta = varargin{1};
end

for i = 1:size(kernels, 2)
    kernel = @(U,V) kernelFn(U, V, kernels{i}, k, lambda, k*theta);
    ticID = tic;
    [stats, ~, classifier] = runTest(training, testing, classes, group, answer, nTrain, nTest, kernel);
    time = toc(ticID);
    results{i}.stats = stats;
    results{i}.classifier = classifier;
    results{i}.time = time;
end

end