/*
 * ssk.cpp
 *
 * Michael Giuffrida - Yale University
 * Spring 2012 - CPSC 490
 *
 * Calculates SSK(s, t, k, lambda) efficiently.
 *
 * This is a MEX-file for MATLAB.
 */

#include <algorithm>
#include "mex.h"

using std::min;
using std::max;

// Matrix used to store previously calculated values of K and K'.
class KMap {
private:
    double* arr;
    size_t m, n, k;
public:
    KMap(size_t m, size_t n, size_t k)
        : m(m), n(n), k(k) {
        size_t length = m*n*k;
        arr = new double[length];
        for (int i = 0; i < length; i++) {
            arr[i] = -1;
        }
    }
    ~KMap() {
        delete[] arr;
    }
    double& at(size_t i, size_t j, size_t p) {
        return arr[(i)*(n*k) + (j)*(k) + p];
    }
};

KMap* kPMap, * kPPMap;
double lambda;
const char* s, * t;
size_t sLen, tLen;
double* lambdas;

double kernel(int k);
double K(size_t endS, size_t endT, unsigned int i);
double KPrime(size_t endS, size_t endT, unsigned int i);
double KPPrime(size_t endS, size_t endT, unsigned int i);
double getLambdaPower(int n);

// Gateway function
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
    // Check for proper number of arguments
    if (nrhs != 4) {
        mexErrMsgIdAndTxt("MATLAB:ssk:nargin", 
            "SSK requires four input arguments.");
    }
    if (nlhs > 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:nargout",
            "SSK requires one output argument.");
    }

    // Check to make sure inputs s, t are row vectors
    if (mxGetM(prhs[0]) != 1 || mxGetNumberOfDimensions(prhs[0]) != 2) {
        mexErrMsgIdAndTxt("MATLAB:ssk:str1NotVector",
            "Input str1 must be a vector.");
    }
    if (mxGetM(prhs[1]) != 1 || mxGetNumberOfDimensions(prhs[1]) != 2) {
        mexErrMsgIdAndTxt("MATLAB:ssk:str1NotVector",
            "Input str2 must be a vector.");
    }

    int k = (int)mxGetScalar(prhs[2]);
    if (k < 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:kTooSmall",
            "Input k must be at least 1.");
    }

    lambda = mxGetScalar(prhs[3]);
    if (lambda < 0 || lambda > 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:lambdaOutOfRange",
            "Input lambda must be between 0 and 1.");
    }
    
    // Copy input strings into vectors of the appropriate type
    s = (const char*)(mxGetData(prhs[0]));
    t = (const char*)(mxGetData(prhs[1]));

    sLen = mxGetN(prhs[0]);
    tLen = mxGetN(prhs[1]);
    
    // Make s the longer string
    if (sLen < tLen) {
        const char* temp = s;
        s = t;
        t = temp;
        size_t tempLen = sLen;
        sLen = tLen;
        tLen = tempLen;
    }
    
    // Return the kernel value as the first element of output matrix
    plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
    mxGetPr(plhs[0])[0] = kernel(k);
}

// Unnomalized kernel
double kernel(int k) {
    // Initialize caches
    kPMap = new KMap(sLen, tLen, k-1);
    kPPMap = new KMap(sLen, tLen, k-1);

    lambdas = new double[max(sLen, tLen)+1];
    lambdas[0] = 1;

    for (int i = 1; i <= max(sLen, tLen); i++) {
        lambdas[i] = lambdas[i-1]*lambda;
    }

    double value = K(sLen-1, tLen-1, k);
    delete kPMap;
    delete kPPMap;
    delete[] lambdas;
    return value;
}

// K_i(sx, t)
double K(size_t endS, size_t endT, unsigned int i) {
    if (min(endS+1, endT+1) < i ||
        endS >= sLen ||
        endT >= tLen) {
        return 0;
    }

    char x = s[endS];
    double value = K(endS-1, endT, i);
    for (unsigned int j = 0; j <= endT; j++) {
        if (t[j] == x) {
            value += KPrime(endS-1, j-1, i-1) * getLambdaPower(2);
        }
    }

    return value;
}

// K'_i(sx,t)
double KPrime(size_t endS, size_t endT, unsigned int i) {
    if (!i) {
        return 1;
    } else if (min(endS+1, endT+1) < i ||
               endS >= sLen ||
               endT >= tLen) {
        return 0;
    }

    double& cached = kPMap->at(endS, endT, i-1);
    if (cached >= 0) {
        return cached;
    }

    char x = s[endS];

    double value = 0;
    value += lambda * KPrime(endS-1, endT, i);
    value += KPPrime(endS, endT, i);

    // Store value in cache map
    cached = value;
    return value;
}

// K''_i(sx, t)
double KPPrime(size_t endS, size_t endT, unsigned int i) {
    if (!i) {
        return 1;
    } else if (min(endS+1, endT+1) < i ||
               endS >= sLen ||
               endT >= tLen) {
        return 0;
    }

    double value = 0;
    char x = s[endS];

    double& cached = kPPMap->at(endS, endT, i-1);
    if (cached >= 0) {
        return cached;
    }

    // Find largest u at end of t such that x does not occur in u
    if (t[endT] == x) {
        // K''_i(sx, tx)
        value = lambda * (KPPrime(endS, endT-1, i) +
                          lambda * KPrime(endS-1, endT-1, i-1));
        cached = value;
        return value;
    }

    size_t found;
    for (found = endT-1; found <= endT; found--) {
        if (t[found] == x) {
            break;
        }
    }

    if (found > endT) {
        // No commonality
        value = 0;
    } else {
        // K''_i(sx, tu)
        value = getLambdaPower(int(endT - found)) *
            KPPrime(endS, found, i);
    }

    cached = value;
    return value;
}

double getLambdaPower(int n) {
    if (n <= max(sLen, tLen)) {
        return lambdas[n];
    } else {
        return lambdas[max(sLen, tLen)] *
            pow(lambda, (int)(n - max(sLen, tLen)));
    }
}
