/*
 * ssk_lp.cpp
 *
 * Michael Giuffrida - Yale University
 * Spring 2012 - CPSC 490
 *
 * Calculates SSK(s, t, k, lambda, theta) efficiently using cached lambda pruning.
 *
 * This is a MEX-file for MATLAB.
 */

#include <algorithm>
#include <cmath>
#include "mex.h"

using std::min;

// Matrix used to store previously calculated values of K and K'.
class KMap {
private:
    double* arr;
    unsigned char* arrM;
    size_t m, n, k;
public:
    KMap(size_t m, size_t n, size_t k)
        : m(m), n(n), k(k) {
        size_t length = m*n*k;
        arr = new double[length];
        arrM = new unsigned char[length]();
    }
    ~KMap() {
        delete[] arr;
        delete[] arrM;
    }
    double& at(size_t i, size_t j, size_t p, int m) {
        size_t index = (i)*(n*k) + (j)*(k) + (p-1);
        // Check m
        if (arrM[index] < m) {
            // Clear out the cached value so it will be replaced.
            arr[index] = -1;
            arrM[index] = m;
        }
        return arr[index];
    }
};

KMap* kPMap;
KMap* kPPMap;
double lambda;
char* s, * t;
size_t sLen, tLen;
double* lambdas;

double kernel(int k, int m);
double K(size_t endS, size_t endT, unsigned int i, int m);
double KPrime(size_t endS, size_t endT, unsigned int i, int m);
double KPPrime(size_t endS, size_t endT, unsigned int i, int m);

// Gateway function
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
    // check for proper number of arguments
    if (nrhs != 5) {
        mexErrMsgIdAndTxt("MATLAB:ssk:nargin", 
            "SSK requires five input arguments.");
    }
    if (nlhs > 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:nargout",
            "SSK requires one output argument.");
    }

    // check to make sure inputs s, t are row vectors
    if (mxGetM(prhs[0]) != 1 || mxGetNumberOfDimensions(prhs[0]) != 2) {
        mexErrMsgIdAndTxt("MATLAB:ssk:str1NotVector",
            "Input str1 must be a vector.");
    }
    if (mxGetM(prhs[1]) != 1 || mxGetNumberOfDimensions(prhs[1]) != 2) {
        mexErrMsgIdAndTxt("MATLAB:ssk:str1NotVector",
            "Input str2 must be a vector.");
    }

    int k = (int)mxGetScalar(prhs[2]);
    if (k < 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:kTooSmall",
            "Input k must be at least 1.");
    }

    lambda = mxGetScalar(prhs[3]);
    if (lambda < 0 || lambda > 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:lambdaOutOfRange",
            "Input lambda must be between 0 and 1.");
    }
    
    int m = (int)mxGetScalar(prhs[4]);
    if (m < 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:mOutOfRange",
            "Input m must be at least 1.");
    }
    
    // Copy input strings into vectors of the appropriate type
    s = (char*)(mxGetData(prhs[0]));
    t = (char*)(mxGetData(prhs[1]));

    sLen = mxGetN(prhs[0]);
    tLen = mxGetN(prhs[1]);

    // Return the kernel value as the first element of output matrix.
    plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
    mxGetPr(plhs[0])[0] = kernel(k, m);
}

// Unnomalized kernel
double kernel(int k, int m) {
    // Initialize caches
    kPMap = new KMap(sLen, tLen, k-1);
    kPPMap = new KMap(sLen, tLen, k-1);

    // Precalculate lambda powers
    lambdas = new double[m];

    lambdas[0] = 1;
    for (int i = 1; i < m; i++) {
        lambdas[i] = lambdas[i-1]*lambda;
    }

    double value = K(sLen-1, tLen-1, k, m);

    delete kPMap;
    delete kPPMap;
    delete[] lambdas;
    return value;
}

// K_i(sx, t)
double K(size_t endS, size_t endT, unsigned int i, int m) {
    if (min(endS+1, endT+1) < i ||
        endS >= sLen ||
        endT >= tLen) {
        return 0;
    } else if (m <= 0 || (unsigned)m < 2*i) {
        return 0;
    }

    double value = 0;
    for (size_t curS = endS; curS > i-1 /*&& curS <= endS*/; curS--) {
        char x = s[curS];
        
        for (size_t j = 1; j <= endT; j++) {
            if (t[j] == x) {
                value += KPrime(curS-1, j-1, i-1, m-2);
            }
        }
    }
    return value * lambdas[2];
}

// K'_i(sx,t)
double KPrime(size_t endS, size_t endT, unsigned int i, int m) {
    if (!i) {
        return 1;
    } else if (min(endS+1, endT+1) < i ||
               endS >= sLen ||
               endT >= tLen) {
        return 0;
    } else if (m <= 0 || (unsigned)m < 2*i) {
        return 0;
    }

    double& cached = kPMap->at(endS, endT, i, m);
    if (cached > -.5) {
        return cached;
    }

    double value = 0;
    
    value += lambda * KPrime(endS-1, endT, i, m-1);
    value += KPPrime(endS, endT, i, m);

    // Store value in cache map
    cached = value;
    return value;
}

// K''_i(sx, t)
double KPPrime(size_t endS, size_t endT, unsigned int i, int m) {
    if (!i) {
        return 1;
    } else if (min(endS+1, endT+1) < i ||
               endS >= sLen ||
               endT >= tLen) {
        return 0;
    } else if (m <= 0 || (unsigned)m < 2*i) {
        return 0;
    }

    double value = 0;
    char x = s[endS];

    double& cached = kPPMap->at(endS, endT, i, m);
    if (cached > -.5) {
        return cached;
    }

    if (t[endT] == x) {
        // K''_i(sx, tx)
        value = lambda * (KPPrime(endS, endT-1, i, m-1) +
                          lambda * KPrime(endS-1, endT-1, i-1, m-2));
        cached = value;
        return value;
    }
    // Find largest u at end of t such that x does not occur in u
    size_t minT = (endT > m) ? endT - m : 0;
    for (size_t j = endT; j >= minT && j <= endT; j--) {
        if (t[j] == x) {
            value = lambdas[endT - j] *
                    KPPrime(endS, j, i, m - (int)(endT - j));
            cached = value;
            return value;
        }
    }
    return 0;
}
