#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include <cmath>
#include <stdint.h>
#include <cstdlib>
#include "mex.h"
#include "matrix.h"

/*
 * sskit.cpp
 *
 * Iterative calculation of K(s1, s2).
 * Do not use!
 *
 * This is a MEX-file for MATLAB.
 */

using namespace std;

// Matrix used to store previously calculated values of K and K'.
class KMap {
private:
    double* arr;
    size_t m, n, k;
public:
    KMap(size_t m, size_t n, size_t k, double v = 0)
        : m(m), n(n), k(k) {
        size_t length = m*n*k;
        arr = (double*)mxMalloc(sizeof(double) * length);
        if (v) {
            for (int i = 0; i < length; i++) {
                arr[i] = v;
            }
        }
    }
    ~KMap() {
        mxFree(arr);
    }
    double& at(size_t i, size_t j, size_t p) {
        return arr[i*(n*k) + j*(k) + p];
    }
};

double kernel(string& s, string& t, int k, double lambda);

// Gateway function
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
    // check for proper number of arguments
    if (nrhs != 4) {
        mexErrMsgIdAndTxt("MATLAB:ssk:nargin", 
            "SSK requires four input arguments.");
    }
    if (nlhs > 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:nargout",
            "SSK requires one output argument.");
    }

    // check to make sure inputs s, t are row vectors
    if (mxGetM(prhs[0]) != 1 || mxGetNumberOfDimensions(prhs[0]) != 2) {
        mexErrMsgIdAndTxt("MATLAB:ssk:str1NotVector",
            "Input str1 must be a vector.");
    }
    if (mxGetM(prhs[1]) != 1 || mxGetNumberOfDimensions(prhs[1]) != 2) {
        mexErrMsgIdAndTxt("MATLAB:ssk:str1NotVector",
            "Input str2 must be a vector.");
    }

    int k = (int)mxGetScalar(prhs[2]);
    if (k < 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:kTooSmall",
            "Input k must be at least 1.");
    }

    double lambda = mxGetScalar(prhs[3]);
    if (lambda < 0 || lambda > 1) {
        mexErrMsgIdAndTxt("MATLAB:ssk:lambdaOutOfRange",
            "Input lambda must be between 0 and 1.");
    }
    
    // Copy input strings into vectors of the appropriate type
    uint8_t* str1 = (uint8_t*)(mxGetData(prhs[0]));
    uint8_t* str2 = (uint8_t*)(mxGetData(prhs[1]));
    vector<uint8_t> v1(mxGetN(prhs[0]));
    vector<uint8_t> v2(mxGetN(prhs[1]));

    copy(str1, str1 + mxGetN(prhs[0]), v1.begin());
    copy(str2, str2 + mxGetN(prhs[1]), v2.begin());

    // Copy vectors into strings
    string s1(v1.begin(), v1.end());
    string s2(v2.begin(), v2.end());

    // Return the kernel value as the first element of output matrix.
    plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
    mxGetPr(plhs[0])[0] = kernel(s1, s2, k, lambda);
}

// Unnormalized kernel
double kernel(string& str1, string& str2, int n, double lambda) {
    size_t len1 = str1.length();
    size_t len2 = str2.length();

    // Initialize caches
    double* kMap = (double*)mxMalloc(sizeof(double) * len1);
    KMap kPMap(len1, len2, n-1, -1);
    KMap kPPMap(len1, len2, n-1, -1);

    for (int i = 0; i < len1; i++) {
        kMap[i] = -1;
    }

    for (int i = 0; i < n-1; i++) {
        for (int s = 0; s < len1; s++) {
            for (int t = 0; t < len2; t++) {
                if (s < i || t <  i) {
                    kPPMap.at(s, t, i) = 0;
                    continue;
                }
                
                char x = str1[s];
                double value;

                // Find the largest u in tu such that x doesn't occur in u
                int found = -1;
                for (int index = t; index >= 0; index--) {
                    if (str2[index] == x) {
                        found = index;
                        break;
                    }
                }

                if (found == -1) {
                    // No commonality
                    value = 0;
                } else if (found == t) {
                    // K''_i(sx, tx)
                    value = 0;

                    if (t > 0) {
                        value = lambda * kPPMap.at(s, t-1, i);
                        if (i > 0) {
                            if (s > 0) {
                                value += pow(lambda, 2) * kPMap.at(s-1, t-1, i-1);
                            }
                        } else {
                            value += pow(lambda, 2);
                        }
                    } else if (i == 0) {
                        value = pow(lambda, 2);
                    }
                } else {
                    // K''_i(sx, tu)
                    value = pow(lambda, (double)(t - found)) * kPPMap.at(s, found, i);
                }
                kPPMap.at(s, t, i) = value;
            }
        }
        
        // Build up kPMap for each letter in s
        for (int s = 0; s < len1; s++) {
            // Build up kPMap for each letter in t
            for (int t = 0; t < len2; t++) {
                // K'_i = 0, if min(|s|, |t|) < i
                if (s < i || t < i) {
                    kPMap.at(s, t, i) = 0;
                    continue;
                }
                
                double value = 0;
                if (s > 0) {
                    value = lambda * kPMap.at(s-1, t, i);
                }
                value += kPPMap.at(s, t, i);
                
                kPMap.at(s, t, i) = value;
            }
        }
        
    }
    
    // Compute K_k(s, t)
    for (int s = 0; s < len1; s++) {
        char x = str1[s];
        double value = 0;
        if (s > 0) {
            value = kMap[s-1];
            
            // Sum for every substring of t ending in tj = x
            for (int t = 0; t < len2; t++) {
                if (str2[t] == x) {
                    double kP = 1;
                    if (n > 1 && t > 0) {
                        kP = kPMap.at(s-1, t-1, n-2);
                    } else {
                        kP = 0;
                    }
                    value += kP * pow(lambda, 2);
                }
            }
        } else {
            for (int t = 0; t < len2; t++) {
                if (str2[t] == x) {
                    double kP = 1;
                    if (n > 1) {
                        kP = 0;
                    }
                    value += kP * pow(lambda, 2);
                }
            }
        }
        kMap[s] = value;
    }
    
    return kMap[len1-1];
}

